from typing import List, Any
import os
from ..core.base import Sink

try:
    import pyarrow as pa
    import pyarrow.parquet as pq
except Exception:
    pa = None
    pq = None


class ParquetSink(Sink):
    def __init__(self, path: str) -> None:
        self.path = path
        self._writer = None

    def _ensure_dir(self) -> None:
        d = os.path.dirname(self.path)
        if d and not os.path.exists(d):
            os.makedirs(d, exist_ok=True)

    def write_batch(self, batch: List[Any]) -> None:
        if not batch:
            return
        if pa is None or pq is None:
            raise RuntimeError("pyarrow not available")
        self._ensure_dir()
        table = pa.Table.from_pylist(batch)
        if self._writer is None:
            self._writer = pq.ParquetWriter(self.path, table.schema)
        self._writer.write_table(table)

    def close(self) -> None:
        if self._writer is not None:
            self._writer.close()
            self._writer = None



from typing import Iterable, Any, List
from abc import ABC, abstractmethod


class Source(ABC):
    @abstractmethod
    def read(self) -> Iterable[Any]:
        raise NotImplementedError

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass


class Transformer(ABC):
    @abstractmethod
    def transform_batch(self, batch: List[Any]) -> List[Any]:
        raise NotImplementedError


class Sink(ABC):
    @abstractmethod
    def write_batch(self, batch: List[Any]) -> None:
        raise NotImplementedError

    def flush(self) -> None:
        pass

    def close(self) -> None:
        pass



from typing import List, Any, Dict
from ..core.base import Transformer


class NormalizeTransformer(Transformer):
    def __init__(self) -> None:
        pass

    def transform_batch(self, batch: List[Any]) -> List[Any]:
        out: List[Any] = []
        for item in batch:
            if isinstance(item, dict):
                out.append({str(k): item[k] for k in item})
            else:
                out.append(item)
        return out



from typing import Iterable, Any
from ..core.base import Source


class BinanceSource(Source):
    def __init__(self, api_key: str | None = None):
        self.api_key = api_key

    def read(self) -> Iterable[Any]:
        raise NotImplementedError("binance source not implemented")



import os
import tempfile
import yaml
from drip_etl.core.engine import run_from_config


class DummyResp:
    def __init__(self, data):
        self._data = data

    def raise_for_status(self):
        return None

    def json(self):
        return self._data


def test_run_from_config_http_csv(monkeypatch):
    def fake_get(url, params=None, headers=None, timeout=30):
        return DummyResp([{"x": 1}, {"x": 2}])

    monkeypatch.setattr("requests.get", fake_get)
    with tempfile.TemporaryDirectory() as d:
        cfg = {
            "source": {"type": "http", "params": {"url": "http://example"}},
            "transformers": [{"type": "normalize"}],
            "sink": {"type": "csv", "params": {"path": os.path.join(d, "out.csv")}},
            "runtime": {"batch_size": 2},
        }
        path = os.path.join(d, "cfg.yaml")
        with open(path, "w", encoding="utf-8") as f:
            yaml.safe_dump(cfg, f)
        run_from_config(path)
        assert os.path.exists(os.path.join(d, "out.csv"))



from typing import List, Any
from ..core.base import Sink


class MemorySink(Sink):
    def __init__(self) -> None:
        self.items: List[Any] = []

    def write_batch(self, batch: List[Any]) -> None:
        self.items.extend(batch)



from drip_etl.transformers.mapping_transformer import MappingTransformer
from drip_etl.transformers.filter_transformer import FilterTransformer
from drip_etl.transformers.calculate_transformer import CalculateTransformer


def test_mapping_transformer():
    t = MappingTransformer(mapping={"x": "a"})
    out = t.transform_batch([{"a": 1}, {"a": 2}])
    assert out == [{"x": 1}, {"x": 2}]


def test_filter_transformer():
    t = FilterTransformer(include_if={"a": 1})
    out = t.transform_batch([{"a": 1}, {"a": 2}])
    assert out == [{"a": 1}]


def test_calculate_transformer():
    t = CalculateTransformer(add={"sum": ["a", "b"]})
    out = t.transform_batch([{"a": 1, "b": 2}])
    assert out[0]["sum"] == 3.0



from typing import List, Any
import csv
import os
from ..core.base import Sink


class CsvSink(Sink):
    def __init__(self, path: str, overwrite: bool = True) -> None:
        self.path = path
        self.overwrite = overwrite
        self._initialized = False
        self._fieldnames: List[str] | None = None

    def _ensure_dir(self) -> None:
        d = os.path.dirname(self.path)
        if d and not os.path.exists(d):
            os.makedirs(d, exist_ok=True)

    def write_batch(self, batch: List[Any]) -> None:
        if not batch:
            return
        self._ensure_dir()
        mode = "w" if self.overwrite and not self._initialized else "a"
        with open(self.path, mode, newline="", encoding="utf-8") as f:
            if isinstance(batch[0], dict):
                if not self._fieldnames:
                    self._fieldnames = list(batch[0].keys())
                writer = csv.DictWriter(f, fieldnames=self._fieldnames)
                if not self._initialized:
                    writer.writeheader()
                    self._initialized = True
                for item in batch:
                    writer.writerow(item)
            else:
                writer = csv.writer(f)
                for item in batch:
                    writer.writerow([item])

    def close(self) -> None:
        pass



import os
import tempfile
from drip_etl.sinks.csv_sink import CsvSink


def test_csv_sink_writes():
    with tempfile.TemporaryDirectory() as d:
        path = os.path.join(d, "out.csv")
        sink = CsvSink(path=path)
        sink.write_batch([{"a": 1}, {"a": 2}])
        sink.close()
        assert os.path.exists(path)
        with open(path, "r", encoding="utf-8") as f:
            content = f.read()
        assert "a" in content



import logging
import os


def get_logger(name: str) -> logging.Logger:
    level_str = os.getenv("DRIP_LOG_LEVEL", "INFO").upper()
    level = getattr(logging, level_str, logging.INFO)
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(level)
        handler = logging.StreamHandler()
        fmt = "%(asctime)s %(levelname)s %(name)s %(message)s"
        handler.setFormatter(logging.Formatter(fmt))
        logger.addHandler(handler)
    return logger



from typing import Iterable, Iterator, List, Any
import time


def now_ms() -> int:
    return int(time.time() * 1000)


def batched(iterable: Iterable[Any], size: int) -> Iterator[List[Any]]:
    batch: List[Any] = []
    for item in iterable:
        batch.append(item)
        if len(batch) >= size:
            yield batch
            batch = []
    if batch:
        yield batch



from typing import Iterable, Any, Dict, Optional
import time
import platform
import pandas as pd
from datetime import datetime
from ..core.base import Source
from ..common.logger import get_logger

logger = get_logger(__name__)


class Mt5StreamSource(Source):
    def __init__(self,
                 symbol: str = "XAUUSD",
                 period: str = "M1",
                 interval: float = 1.0,
                 limit: int = 10,
                 login: int = 0,
                 password: str = "",
                 server: str = "") -> None:
        self.symbols = [s.strip() for s in symbol.split(',')]
        self.periods = [p.strip() for p in period.split(',')]
        self.interval = interval
        self.limit = limit
        self.login = login
        self.password = password
        self.server = server
        self.platform = platform.system()
        # Key: (symbol, period) -> {time: record}
        self.last_records: Dict[tuple, Dict[Any, Dict[str, Any]]] = {}
        # Key: (symbol, period) -> latest_time
        self.last_seen_max_times: Dict[tuple, Any] = {}
        
        if self.limit < 2:
            logger.warning("Limit < 2 might miss candle completion updates. Setting limit to 2.")
            self.limit = 2
        
        if self.platform != "Windows":
             logger.warning("MT5 source usually requires Windows to run the terminal.")

    def read(self) -> Iterable[Dict[str, Any]]:
        try:
            # Delayed import to avoid hard dependency on non-Windows
            from ..common.mt5_conn import MT5Connector, MT5Config, MT5Period
        except ImportError as e:
            logger.error(f"Required dependencies for MT5 not found: {e}. Please install 'drip-etl[mt5]'.")
            raise

        config = MT5Config(
            login=self.login,
            password=self.password,
            server=self.server
        )

        try:
            with MT5Connector(config) as conn:
                # Prepare period mappings
                period_map = {}
                for p in self.periods:
                    try:
                        period_enum = getattr(MT5Period, p)
                        period_map[p] = period_enum.value
                    except AttributeError:
                        logger.warning(f"Unknown period {p}, skipping")

                if not period_map:
                    logger.error("No valid periods found. Defaulting to M1.")
                    period_map['M1'] = MT5Period.M1.value
                    self.periods = ['M1']

                logger.info(f"Starting MT5 stream for {self.symbols} {self.periods} (interval={self.interval}s)")

                while True:
                    for symbol in self.symbols:
                        for period_name in self.periods:
                            period_value = period_map.get(period_name)
                            if period_value is None:
                                continue
                                
                            key = (symbol, period_name)
                            if key not in self.last_records:
                                self.last_records[key] = {}
                            
                            try:
                                # Fetch latest data
                                df = conn.get_kline_data(symbol, period=period_value, count=self.limit)
                                
                                if df is not None and not df.empty:
                                    # Convert time to datetime if it's integer timestamp
                                    if pd.api.types.is_integer_dtype(df['time']):
                                        df['time'] = pd.to_datetime(df['time'], unit='s')
                                    
                                    records = df.to_dict('records')
                                    new_records_count = 0
                                    
                                    latest_time = df['time'].max()
                                    last_seen_max_time = self.last_seen_max_times.get(key)
                                    
                                    # Detect new period start
                                    if last_seen_max_time is not None and latest_time > last_seen_max_time:
                                        logger.info(f"New period detected for {symbol} {period_name}: {latest_time} (was {last_seen_max_time}). Finalizing previous candle.")
                                    
                                    self.last_seen_max_times[key] = latest_time

                                    for record in records:
                                        # Inject symbol and period if missing
                                        if 'symbol' not in record:
                                            record['symbol'] = symbol
                                        if 'period' not in record:
                                            record['period'] = period_name
                                        
                                        record_time = record['time']
                                        
                                        # Add is_completed flag
                                        record['is_completed'] = 1 if record_time < latest_time else 0

                                        # Check if record is new or changed
                                        should_yield = False
                                        
                                        if record_time not in self.last_records[key]:
                                            should_yield = True
                                        else:
                                            last_record = self.last_records[key][record_time]
                                            if record != last_record:
                                                should_yield = True
                                        
                                        if should_yield:
                                            self.last_records[key][record_time] = record.copy()
                                            yield record
                                            new_records_count += 1
                                    
                                    if new_records_count > 0:
                                        logger.debug(f"Yielded {new_records_count} updated records for {symbol} {period_name}")
                                        
                                    # Prune cache
                                    if len(self.last_records[key]) > self.limit * 2:
                                        sorted_times = sorted(self.last_records[key].keys())
                                        keys_to_remove = sorted_times[:-self.limit]
                                        for k in keys_to_remove:
                                            del self.last_records[key][k]
                                
                                else:
                                    logger.warning(f"No data received for {symbol} {period_name}")

                            except Exception as e:
                                logger.error(f"Error in stream loop for {symbol} {period_name}: {e}")
                    
                    time.sleep(self.interval)

        except Exception as e:
            logger.error(f"Fatal error connecting to MT5: {e}")
            raise



from typing import Iterable, Any
from ..core.base import Source


class SimpleListSource(Source):
    def __init__(self, data):
        self.data = data

    def read(self) -> Iterable[Any]:
        for x in self.data:
            yield x



from typing import Iterable, Any, List
from .base import Source, Transformer, Sink
from ..common.logger import get_logger
from ..common.metrics import Metrics
from ..common.utils import batched


class Pipeline:
    def __init__(self, source: Source, transformers: List[Transformer], sink: Sink, batch_size: int = 500) -> None:
        self.source = source
        self.transformers = transformers
        self.sink = sink
        self.batch_size = batch_size
        self.logger = get_logger("drip_etl.pipeline")
        self.metrics = Metrics()

    def run(self) -> None:
        self.logger.info("pipeline_start")
        self.source.start()
        try:
            for batch in batched(self.source.read(), self.batch_size):
                out = batch
                for t in self.transformers:
                    out = t.transform_batch(out)
                if out:
                    self.sink.write_batch(out)
                    self.metrics.inc("batches_written", 1)
                    self.metrics.inc("records_written", len(out))
            self.sink.flush()
        finally:
            self.source.stop()
            self.sink.close()
        self.logger.info("pipeline_done")



from typing import Any, Dict
import os
import yaml


def _env_substitute(value: Any) -> Any:
    if isinstance(value, str) and value.startswith("${ENV:") and value.endswith("}"):
        key = value[6:-1]
        return os.getenv(key, "")
    return value


def _walk_env(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {k: _walk_env(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_walk_env(v) for v in obj]
    return _env_substitute(obj)


def load_config(path: str) -> Dict[str, Any]:
    with open(path, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f) or {}
    if "source" not in data or "sink" not in data:
        raise ValueError("config must include source and sink")
    if "type" not in data["source"] or "type" not in data["sink"]:
        raise ValueError("source/sink must include type")
    data = _walk_env(data)
    return data



from .normalize_transformer import NormalizeTransformer
from .mapping_transformer import MappingTransformer
from .filter_transformer import FilterTransformer
from .calculate_transformer import CalculateTransformer



from typing import List, Any, Dict
from ..core.base import Transformer


class MappingTransformer(Transformer):
    def __init__(self, mapping: Dict[str, str] | None = None) -> None:
        self.mapping = mapping or {}

    def transform_batch(self, batch: List[Any]) -> List[Any]:
        out: List[Any] = []
        for item in batch:
            if isinstance(item, dict) and self.mapping:
                mapped = {}
                for new_key, old_key in self.mapping.items():
                    mapped[new_key] = item.get(old_key)
                out.append(mapped)
            else:
                out.append(item)
        return out



import MetaTrader5 as mt5
import pandas as pd
from enum import Enum


class MT5Period(Enum):
    M1 = mt5.TIMEFRAME_M1
    M2 = mt5.TIMEFRAME_M2
    M3 = mt5.TIMEFRAME_M3
    M4 = mt5.TIMEFRAME_M4
    M5 = mt5.TIMEFRAME_M5
    M6 = mt5.TIMEFRAME_M6
    M10 = mt5.TIMEFRAME_M10
    M12 = mt5.TIMEFRAME_M12
    M15 = mt5.TIMEFRAME_M15
    M20 = mt5.TIMEFRAME_M20
    M30 = mt5.TIMEFRAME_M30
    H1 = mt5.TIMEFRAME_H1
    H2 = mt5.TIMEFRAME_H2
    H3 = mt5.TIMEFRAME_H3
    H4 = mt5.TIMEFRAME_H4
    H6 = mt5.TIMEFRAME_H6
    H8 = mt5.TIMEFRAME_H8
    H12 = mt5.TIMEFRAME_H12
    D1 = mt5.TIMEFRAME_D1
    W1 = mt5.TIMEFRAME_W1
    MN1 = mt5.TIMEFRAME_MN1
    
    @classmethod
    def get_name(cls, value):
        try:
            return cls(value).name
        except ValueError:
            return str(value)

class MT5Config:
    def __init__(self, login,password,server):
        self.login = login
        self.password = password
        self.server = server

class MT5Connector:
    
    def __init__(self, config: MT5Config):
        self.config = config
        self.login = config.login
        self.password = config.password
        self.server = config.server
        self.is_connected = False
        self.is_authorized = False


    def connect(self):
        try:
            if not mt5.initialize():
                print("MT5初始化失败")
                return False
            
            self.is_connected = True
            
            if self.login and self.password:
                return self.login_account()
            else:
                return True
                
        except Exception as e:
            print(f"连接MT5失败: {e}")
            return False
    
    def login_account(self):
        try:
            authorized = mt5.login(self.login, self.password, self.server)
            if authorized:
                self.is_authorized = True
                return True
            else:
                print(f"账户登录失败: {mt5.last_error()}")
                return False
        except Exception as e:
            print(f"登录失败: {e}")
            return False
    
    def disconnect(self):
        try:
            mt5.shutdown()
            self.is_connected = False
            self.is_authorized = False
        except Exception as e:
            print(f"断开连接失败: {e}")
    
    def get_symbols(self, limit=10):
        if not self.is_connected:
            print("未连接到MT5")
            return []
        
        try:
            symbols = mt5.symbols_get()
            return symbols
        except Exception as e:
            print(f"获取交易品种失败: {e}")
            return []
    
    def get_symbol_info(self, symbol):
        if not self.is_connected:
            print("未连接到MT5")
            return None
        
        try:
            symbol_info = mt5.symbol_info(symbol)
            if symbol_info is not None:
                print(f"{symbol}点差: {symbol_info.spread}")
                print(f"合约大小: {symbol_info.trade_contract_size}")
                return symbol_info
            else:
                print(f"未找到品种: {symbol}")
                return None
        except Exception as e:
            print(f"获取品种信息失败: {e}")
            return None
    
    def enable_symbols(self, symbols):
        if not self.is_connected:
            print("未连接到MT5")
            return False
        
        try:
            enabled_count = 0
            for symbol in symbols:
                symbol_info = mt5.symbol_info(symbol)
                if symbol_info is None or not symbol_info.visible:
                    if mt5.symbol_select(symbol, True):
                        enabled_count += 1
                    else:
                        print(f"启用品种失败: {symbol}")
                else:
                    enabled_count += 1

            print(f"成功启用 {enabled_count}/{len(symbols)} 个品种")
            return enabled_count == len(symbols)
        except Exception as e:
            print(f"启用品种失败: {e}")
            return False
    
    def get_symbol_tick(self, symbols):
        if not self.is_connected:
            print("未连接到MT5")
            return {}
        
        try:
            rates = {}
            for symbol in symbols:
                tick = mt5.symbol_info_tick(symbol)
                
                if tick is not None:
                    rates[symbol] = {
                        'bid': tick.bid,
                        'ask': tick.ask,
                        'last': tick.last,
                        'volume': tick.volume,
                        'time': pd.to_datetime(tick.time, unit='s'),
                        'spread': tick.ask - tick.bid
                    }
                else:
                    print(f"获取 {symbol} 报价失败")
            return rates
        except Exception as e:
            print(f"获取实时报价失败: {e}")
            return {}
    
    def get_kline_data(self, symbol, period, count=1000, start_pos=0):
        if not self.is_connected:
            print("未连接到MT5")
            return None
        
        try:
            rates = mt5.copy_rates_from_pos(symbol, period, start_pos, count)
            
            if rates is None or len(rates) == 0:
                print(f"获取 {symbol} K线数据失败 (返回为空)")
                return None
                
            df = pd.DataFrame(rates)
            df['time'] = pd.to_datetime(df['time'], unit='s')
            df['period'] = MT5Period.get_name(period)
            
            return df
        except Exception as e:
            print(f"获取K线数据异常: {e}")
            return None

    def get_kline_range(self, symbol, period, date_from, date_to):
        if not self.is_connected:
            print("未连接到MT5")
            return None
            
        try:
            rates = mt5.copy_rates_range(symbol, period, date_from, date_to)
            
            if rates is None or len(rates) == 0:
                print(f"获取 {symbol} 范围K线数据失败")
                return None
                
            df = pd.DataFrame(rates)
            df['time'] = pd.to_datetime(df['time'], unit='s')
            df['period'] = MT5Period.get_name(period)
            return df
        except Exception as e:
            print(f"获取范围K线数据异常: {e}")
            return None
    
    def __enter__(self):
        self.connect()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()


def main():
    with MT5Connector() as conn:
        symbols = ["EURUSD", "GBPUSD", "XAUUSD", "BTCUSD", "ETHUSD"]
        
        conn.enable_symbols(symbols)
        
        rates = conn.get_symbol_tick(symbols)
        
        if rates:
            df = pd.DataFrame(rates).T
            print("\n实时报价汇总:")
            print(df)
            
        print("\n获取 XAUUSD 最近10根M1 K线:")
        klines = conn.get_kline_data("XAUUSD", period=mt5.TIMEFRAME_M1, count=10)
        if klines is not None:
            print(klines)

if __name__ == '__main__':
    main()



from typing import List, Any
from ..core.base import Sink


class MySqlSink(Sink):
    def __init__(self, dsn: str, table: str) -> None:
        self.dsn = dsn
        self.table = table

    def write_batch(self, batch: List[Any]) -> None:
        raise NotImplementedError("mysql sink not implemented")



from drip_etl.core.pipeline import Pipeline
from drip_etl.transformers.base import PassThroughTransformer
from drip_etl.sources.base import SimpleListSource
from drip_etl.sinks.base import MemorySink


def test_pipeline_runs():
    src = SimpleListSource([{"a": 1}, {"a": 2}, {"a": 3}])
    t = PassThroughTransformer()
    sink = MemorySink()
    p = Pipeline(src, [t], sink, batch_size=2)
    p.run()
    assert len(sink.items) == 3



from typing import Iterable, Any, Dict
import requests
from ..core.base import Source
from ..common.logger import get_logger


class HttpSource(Source):
    def __init__(self, url: str, method: str = "GET", params: Dict[str, Any] | None = None, headers: Dict[str, str] | None = None):
        self.url = url
        self.method = method.upper()
        self.params = params or {}
        self.headers = headers or {}
        self.logger = get_logger("drip_etl.http_source")

    def read(self) -> Iterable[Any]:
        if self.method == "GET":
            resp = requests.get(self.url, params=self.params, headers=self.headers, timeout=30)
        else:
            resp = requests.post(self.url, json=self.params, headers=self.headers, timeout=30)
        resp.raise_for_status()
        data = resp.json()
        if isinstance(data, list):
            for x in data:
                yield x
        elif isinstance(data, dict):
            if "data" in data and isinstance(data["data"], list):
                for x in data["data"]:
                    yield x
            else:
                yield data
        else:
            yield data



from typing import Iterable, Any
from ..core.base import Source


class WebSocketSource(Source):
    def __init__(self, url: str, topics: list | None = None):
        self.url = url
        self.topics = topics or []

    def read(self) -> Iterable[Any]:
        raise NotImplementedError("websocket source not implemented")



from typing import List, Any
from ..core.base import Transformer


class FilterTransformer(Transformer):
    def __init__(self, include_if: dict | None = None, exclude_if: dict | None = None) -> None:
        self.include_if = include_if or {}
        self.exclude_if = exclude_if or {}

    def _match(self, item: Any, cond: dict) -> bool:
        if not isinstance(item, dict):
            return False
        for k, v in cond.items():
            if item.get(k) != v:
                return False
        return True

    def transform_batch(self, batch: List[Any]) -> List[Any]:
        out: List[Any] = []
        for item in batch:
            if self.include_if and not self._match(item, self.include_if):
                continue
            if self.exclude_if and self._match(item, self.exclude_if):
                continue
            out.append(item)
        return out



from typing import Iterable, Any, Dict, Optional
from datetime import datetime
from ..core.base import Source
from ..common.logger import get_logger
import platform
import pandas as pd

logger = get_logger(__name__)


class Mt5Source(Source):
    def __init__(self,
                 symbol: str = "XAUUSD",
                 period: str = "M1",
                 start_time: Optional[str] = None,
                 end_time: Optional[str] = None,
                 limit: int = 100,
                 login: int = 0,
                 password: str = "",
                 server: str = "") -> None:
        self.symbol = symbol
        self.period = period
        self.start_time = start_time
        self.end_time = end_time
        self.limit = limit
        self.login = login
        self.password = password
        self.server = server
        self.platform = platform.system()
        
        if self.platform != "Windows":
             logger.warning("MT5 source usually requires Windows to run the terminal.")

    def read(self) -> Iterable[Dict[str, Any]]:
        try:
            # Delayed import to avoid hard dependency on non-Windows
            from ..common.mt5_conn import MT5Connector, MT5Config, MT5Period
        except ImportError as e:
            logger.error(f"Required dependencies for MT5 not found: {e}. Please install 'drip-etl[mt5]'.")
            raise

        config = MT5Config(
            login=self.login,
            password=self.password,
            server=self.server
        )

        try:
            with MT5Connector(config) as conn:
                # Map period string to MT5Period enum
                try:
                    period_enum = getattr(MT5Period, self.period)
                    period_value = period_enum.value
                except AttributeError:
                    logger.warning(f"Unknown period {self.period}, defaulting to M1")
                    from ..common.mt5_conn import MT5Period # ensure it's available
                    period_value = MT5Period.M1.value

                if self.start_time:
                    # Range mode
                    try:
                        date_from = datetime.fromisoformat(self.start_time)
                        now = datetime.now()
                        if self.end_time:
                            date_to = min(datetime.fromisoformat(self.end_time), now)
                        else:
                            date_to = now
                            
                        df = conn.get_kline_range(self.symbol, period=period_value, date_from=date_from, date_to=date_to)
                    except ValueError as e:
                        logger.error(f"Invalid date format (use 'YYYY-MM-DD HH:MM:SS'): {e}")
                        raise
                else:
                    # Count mode (fallback)
                    df = conn.get_kline_data(self.symbol, period=period_value, count=self.limit)
                
                if df is not None and not df.empty:
                    logger.info(f"Retrieved {len(df)} records from MT5 for {self.symbol}")
                    # Convert time to datetime if it's integer timestamp
                    if pd.api.types.is_integer_dtype(df['time']):
                        df['time'] = pd.to_datetime(df['time'], unit='s')
                    
                    records = df.to_dict('records')
                    for record in records:
                        # Inject period if missing
                        if 'period' not in record:
                            record['period'] = self.period
                        yield record
                else:
                    logger.warning(f"No data received for {self.symbol}")

        except Exception as e:
            logger.error(f"Error reading from MT5: {e}")
            raise



import argparse
from typing import Dict, Any, List
from importlib import import_module
from .pipeline import Pipeline
from .base import Source, Transformer, Sink
from .config_loader import load_config
from ..common.logger import get_logger


def build_component(type_name: str, params: Dict[str, Any], kind: str):
    mapping = {
        "source": {
            "http": "drip_etl.sources.http_source:HttpSource",
            "mt5": "drip_etl.sources.mt5_source:Mt5Source",
            "mt5stream": "drip_etl.sources.mt5stream_source:Mt5StreamSource",
            "okx": "drip_etl.sources.okx_source:OkxSource",
            "binance": "drip_etl.sources.binance_source:BinanceSource",
            "websocket": "drip_etl.sources.websocket_source:WebsocketSource",
        },
        "transformer": {
            "normalize": "drip_etl.transformers.normalize_transformer:NormalizeTransformer",
            "mapping": "drip_etl.transformers.mapping_transformer:MappingTransformer",
            "filter": "drip_etl.transformers.filter_transformer:FilterTransformer",
            "calculate": "drip_etl.transformers.calculate_transformer:CalculateTransformer",
        },
        "sink": {
            "csv": "drip_etl.sinks.csv_sink:CsvSink",
            "parquet": "drip_etl.sinks.parquet_sink:ParquetSink",
            "postgresql": "drip_etl.sinks.postgresql_sink:PostgreSqlSink",
            "mysql": "drip_etl.sinks.mysql_sink:MysqlSink",
        },
    }
    target = mapping[kind].get(type_name)
    if not target:
        raise ValueError(f"unknown {kind} type: {type_name}")
    module_name, cls_name = target.split(":")
    module = import_module(module_name)
    cls = getattr(module, cls_name)
    return cls(**(params or {}))


def run_from_dict(cfg: Dict[str, Any]) -> None:
    logger = get_logger("drip_etl.engine")
    source_cfg = cfg["source"]
    t_cfgs: List[Dict[str, Any]] = cfg.get("transformers", [])
    sink_cfg = cfg["sink"]
    runtime = cfg.get("runtime", {})
    source: Source = build_component(source_cfg["type"], source_cfg.get("params", {}), "source")
    transformers: List[Transformer] = []
    for tc in t_cfgs:
        transformers.append(build_component(tc["type"], tc.get("params", {}), "transformer"))
    sink: Sink = build_component(sink_cfg["type"], sink_cfg.get("params", {}), "sink")
    batch_size = int(runtime.get("batch_size", 500))
    p = Pipeline(source, transformers, sink, batch_size=batch_size)
    logger.info("run_from_config")
    p.run()


def run_from_config(path: str) -> None:
    cfg = load_config(path)
    run_from_dict(cfg)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", required=True)
    args = parser.parse_args()
    run_from_config(args.config)



from typing import List, Any, Dict, Optional
from urllib.parse import quote_plus
from ..core.base import Sink
from ..common.logger import get_logger
from sqlalchemy import create_engine, text
from sqlalchemy.dialects.postgresql import insert
import pandas as pd

logger = get_logger(__name__)


class PostgreSqlSink(Sink):
    def __init__(self, table: str, 
                 dsn: Optional[str] = None,
                 host: Optional[str] = None, 
                 port: int = 5432, 
                 user: Optional[str] = None, 
                 password: Optional[str] = None, 
                 database: Optional[str] = None,
                 upsert_keys: Optional[List[str]] = None,
                 batch_size: int = 1000) -> None:
        self.table = table
        self.upsert_keys = upsert_keys
        self.batch_size = batch_size
        
        if dsn:
            self.dsn = dsn
        elif host and user and database:
            # Construct DSN with URL-encoded password
            pwd = quote_plus(password) if password else ""
            self.dsn = f"postgresql+psycopg2://{user}:{pwd}@{host}:{port}/{database}"
        else:
            raise ValueError("Must provide either 'dsn' or connection parameters (host, user, password, database)")
            
        self.engine = None
        self._init_engine()

    def _init_engine(self):
        try:
            self.engine = create_engine(self.dsn)
            # Verify connection
            with self.engine.connect() as conn:
                pass
            # Log connection info without exposing credentials
            if "@" in self.dsn:
                host_info = self.dsn.split('@')[-1]
                logger.info(f"Connected to PostgreSQL: {host_info}")
            else:
                logger.info("Connected to PostgreSQL (DSN hidden)")
        except Exception as e:
            logger.error(f"Failed to connect to PostgreSQL: {e}")
            raise

    def _create_upsert_method(self, upsert_keys: List[str]):
        def method(table, conn, keys, data_iter):
            # Convert data iterator to list of dicts
            data = [dict(zip(keys, row)) for row in data_iter]
            
            if not data:
                return

            # Create insert statement
            stmt = insert(table.table).values(data)
            
            # Construct set_ dict: update all columns that are being inserted, except the conflict keys
            set_dict = {
                key: getattr(stmt.excluded, key)
                for key in keys
                if key not in upsert_keys
            }
            
            # Create upsert statement
            upsert_stmt = stmt.on_conflict_do_update(
                index_elements=upsert_keys,
                set_=set_dict
            )
            
            # Execute statement
            conn.execute(upsert_stmt)
            
        return method

    def write_batch(self, batch: List[Dict[str, Any]]) -> None:
        if not batch:
            return
            
        try:
            # Debug log to inspect data
            if logger.isEnabledFor(10): # DEBUG level
                logger.debug(f"Writing batch of {len(batch)} records. First record: {batch[0]}")
                
            df = pd.DataFrame(batch)
            # Use 'append' to add to existing table
            # Use connection context for better transaction handling and compatibility
            with self.engine.begin() as connection:
                if self.upsert_keys:
                    method = self._create_upsert_method(self.upsert_keys)
                    df.to_sql(self.table, connection, if_exists='append', index=False, method=method)
                else:
                    df.to_sql(self.table, connection, if_exists='append', index=False, method='multi')
            logger.info(f"Wrote {len(batch)} records to table {self.table}")
        except Exception as e:
            logger.error(f"Failed to write batch to PostgreSQL: {e}")
            raise

    def close(self) -> None:
        if self.engine:
            self.engine.dispose()



from setuptools import setup, find_packages


setup(
    name="drip-etl",
    version="0.1.7",
    description="Modular ETL pipeline driven by YAML configuration",
    author="luoji2026",
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.10",
    install_requires=[
        "PyYAML>=6.0",
        "requests>=2.31.0",
        "pyarrow>=14.0.2",
    ],
    extras_require={
        "exchanges": ["ccxt>=4.0.0"],
        "db": ["sqlalchemy>=2.0.0", "psycopg2-binary>=2.9.0", "pymysql>=1.0.0"],
        "mt5": ["MetaTrader5>=5.0.0", "pandas>=2.0.0"],
    },
    entry_points={
        "console_scripts": [
            "drip-etl=drip_etl.core.engine:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license="MIT",
    license_files=[],
)



from drip_etl.sources.http_source import HttpSource


class DummyResp:
    def __init__(self, data):
        self._data = data

    def raise_for_status(self):
        return None

    def json(self):
        return self._data


def test_http_source_list(monkeypatch):
    def fake_get(url, params=None, headers=None, timeout=30):
        return DummyResp([{"a": 1}, {"a": 2}])

    monkeypatch.setattr("requests.get", fake_get)
    src = HttpSource(url="http://x")
    out = list(src.read())
    assert len(out) == 2



import time
from typing import Dict, Any


class Metrics:
    def __init__(self) -> None:
        self.counters: Dict[str, int] = {}
        self.gauges: Dict[str, float] = {}
        self.histograms: Dict[str, list] = {}
        self.timestamps: Dict[str, float] = {}

    def inc(self, name: str, value: int = 1) -> None:
        self.counters[name] = self.counters.get(name, 0) + value

    def set(self, name: str, value: float) -> None:
        self.gauges[name] = value

    def observe(self, name: str, value: float) -> None:
        arr = self.histograms.get(name)
        if arr is None:
            arr = []
            self.histograms[name] = arr
        arr.append(value)

    def mark(self, name: str) -> None:
        self.timestamps[name] = time.time()

    def snapshot(self) -> Dict[str, Any]:
        return {
            "counters": dict(self.counters),
            "gauges": dict(self.gauges),
            "histograms": {k: list(v) for k, v in self.histograms.items()},
            "timestamps": dict(self.timestamps),
        }



from typing import List, Any, Dict
from ..core.base import Transformer


class CalculateTransformer(Transformer):
    def __init__(self, add: Dict[str, List[str]] | None = None, calculations: List[Dict[str, str]] | None = None) -> None:
        self.add = add or {}
        self.calculations = calculations or []

    def transform_batch(self, batch: List[Any]) -> List[Any]:
        out: List[Any] = []
        for item in batch:
            if isinstance(item, dict):
                # Legacy add support
                if self.add:
                    for new_key, keys in self.add.items():
                        total = 0.0
                        for k in keys:
                            v = item.get(k)
                            if isinstance(v, (int, float)):
                                total += float(v)
                        item[new_key] = total
                
                # New calculations support
                if self.calculations:
                    for calc in self.calculations:
                        field = calc.get("field")
                        expression = calc.get("expression")
                        if field and expression:
                            try:
                                # Use item as context for evaluation
                                item[field] = eval(expression, {"__builtins__": {}}, item)
                            except Exception:
                                # Ignore calculation errors (e.g. missing fields)
                                pass
                out.append(item)
            else:
                out.append(item)
        return out
